

/*  This code performs NITER iterations of an NxN life board.  */

#define N             64
#define NITER         50

#define NNODES        4

#include <stdio.h>


int node(int, int) ;

void real_main()
{
  int blockSizeMax = (N + NNODES - 1) / NNODES ;
  int (*block) [N] ;

  int blockBase, blockSize ;

  int iter ;

  /*  Create the node processes.  */

  blockBase = 0 ;
  while(blockBase < N) {
    if(blockBase + blockSizeMax > N)
      blockSize = N - blockBase ;
    else
      blockSize = blockSizeMax ;

    eval("worker", node(blockBase, blockSize)) ;

    blockBase += blockSize ;
  }

  /*  Handle output.  */

  block = (int (*) [N]) malloc(blockSizeMax * sizeof(int [N])) ;

  for(iter = 0 ; iter <= NITER ; iter++) {

    /*  Copy current state of board from tuple space to a `.pgm' file.  */

    char fname [256] ;
    FILE* f ;
    int i, y ;
 
    sprintf(fname, "life%2.2d.pgm", iter) ;

    f = fopen(fname, "w") ;
    fprintf(f, "P2\n%3d  %3d\n%3d\n", N, N, 1) ;

    blockBase = 0 ;
    while(blockBase < N) {
      in("block", iter, blockBase, ? block : blockSize) ;
    
      for(i = 0 ; i < blockSize ; i++) {
        for(y = 0 ; y < N ; y++)
          fprintf(f, "%d ", block [i] [y]) ;
        fprintf(f, "\n") ;
      }

      blockBase += blockSize ;
    }

    fclose(f) ;
  }
}


int node(int blockBase, int blockSize) {

  int (*block) [N] ;
  int (*neighbours) [N] ;

  int i, y ;
  int iter ;

  /*  `block' has `blockSize + 2' columns and is offset so that
      subscripts start from -1.  This allows for ghost cells.  */

  block = (int (*) [N]) malloc((blockSize + 2) * sizeof(int [N])) + 1 ;

  /*  Initialise board.  */

  for(i = 0 ; i < blockSize ; i++) {
    for(y = 0 ; y < N ; y++) {
      int x = blockBase + i ;
      if(x == N / 2 || y == N / 2)
        block [i] [y] = 1 ;
      else
        block [i] [y] = 0 ;
    }
  }

  /*  Dump initial state of board to tuple space (for host) */

  out("block", 0, blockBase, block : blockSize) ;

  /*  Main update loop.  */

  neighbours = (int (*) [N]) malloc(blockSize * sizeof(int [N])) ;

  for(iter = 1 ; iter <= NITER ; iter++) {

    /*  Put this block's edges into tuple space.  */

    out("edge", iter, blockBase,                 block [0]) ;
    out("edge", iter, blockBase + blockSize - 1, block [blockSize - 1]) ;

    /*  Get ghost edges from tuple space.  */

    in("edge", iter, (blockBase - 1 + N) % N,     ? block [-1]) ;
    in("edge", iter, (blockBase + blockSize) % N, ? block [blockSize]) ;

    /*  Calculate a block of neighbour sums.  */

    for(i = 0 ; i < blockSize ; i++) {
      for(y = 0 ; y < N ; y++) {
        int y_n = (y - 1 + N) % N ;
        int y_p = (y + 1) % N ;

        neighbours [i] [y] =
             block [i - 1] [y_n] + block [i - 1] [y] + block [i - 1] [y_p] +
             block [i]     [y_n]                     + block [i]     [y_p] +
             block [i + 1] [y_n] + block [i + 1] [y] + block [i + 1] [y_p] ;
      }
    }

    /*  Update block of board values.  */

    for(i = 0 ; i < blockSize ; i++) {
      for(y = 0 ; y < N ; y++) {
        int neighbour = neighbours [i] [y] ;

        if(neighbour < 2 || neighbour > 3)
          block [i] [y] = 0 ;
        if(neighbour == 3)
          block [i] [y] = 1 ;
      }
    }

    /*  Dump current state of board to tuple space (for host) */

    out("block", iter, blockBase, block : blockSize) ;
  }

  return 0 ;
}

